#include <cmath>
#include <cstdlib>

#include <pico/stdio_usb.h>
#include <pico/stdlib.h>
#include <pico/time.h>
#include <hardware/pwm.h>

#include <picoled/SSD1306.h>

namespace
{

struct rect
{
  float x, y, w, h;

  void draw(picoled::oled& oled) const
  {
    for (int i = 0; i < w; i++)
    {
      for (int j = 0; j < h; j++)
      {
        oled.write(x + i, y + j, 0xFF);
      }
    }
  }

  float left() const { return x; }
  float right() const { return x + w; }
  float top() const { return y; }
  float bottom() const { return y + h; }

  float middle_h() const { return x + w / 2.0f; }
  float middle_v() const { return y + h / 2.0f; }

  bool overlaps(const rect& other) const
  {
    const bool x_overlaps =
      (left() >= other.left() && left() <= other.right()) ||
      (right() >= other.left() && right() <= other.right());

    const bool y_overlaps =
      (top() >= other.top() && top() <= other.bottom()) ||
      (bottom() >= other.top() && bottom() <= other.bottom());

    return x_overlaps && y_overlaps;
  }
};

int32_t get_encoder_delta(i2c_inst* i2c, uint8_t address)
{
  // SEESAW_ENCODER_BASE and SEESAW_ENCODER_DELTA values
  // https://github.com/adafruit/Adafruit_Seesaw/blob/master/Adafruit_seesaw.h
  uint8_t buf[4] = { 0x11, 0x40, 0x00, 0x00 };
  i2c_write_blocking(i2c, address, buf, 2, false);
  sleep_us(250);
  i2c_read_blocking(i2c, address, buf, 4, false);
  return
    (static_cast<int32_t>(buf[0]) << 24) |
    (static_cast<int32_t>(buf[1]) << 16) |
    (static_cast<int32_t>(buf[2]) << 8 ) |
    (static_cast<int32_t>(buf[3]) << 0 );
}

constexpr uint8_t LEFT_ENCODER_ADDRESS = 0x36;
constexpr uint8_t RIGHT_ENCODER_ADDRESS = 0x37;
constexpr uint8_t OLED_ADDRESS = 0x3C;

constexpr float PADDLE_SPEED = 20;
constexpr float INITIAL_BALL_SPEED = 32;

constexpr uint8_t BUZZER_PIN = 2;

int64_t unbeep(alarm_id_t, void*)
{
  pwm_set_enabled(pwm_gpio_to_slice_num(BUZZER_PIN), false);
  return 0;
}

void beep()
{
  pwm_set_enabled(pwm_gpio_to_slice_num(BUZZER_PIN), true);

  add_alarm_in_ms(40, unbeep, nullptr, false);
}

}

int main()
{
  stdio_usb_init();
  srand(time_us_64());

  i2c_init(i2c0, 1000 * 1000);
  gpio_set_function(0, GPIO_FUNC_I2C);
  gpio_set_function(1, GPIO_FUNC_I2C);
  gpio_pull_up(0);
  gpio_pull_up(1);

  picoled::ssd1306 oled(i2c0, OLED_ADDRESS);
  oled.update();

  rect player = { 8.0f, oled.get_height() / 2.0f - 8.0f, 4.0f, 16.0f };
  rect ai = { oled.get_width() - 8.0f - 4.0f, oled.get_height() / 2.0f - 8.0f, 4.0f, 16.0f };
  rect ball = { 64.0f - 2.0f, 32.0f - 2.0f, 4.0f, 4.0f };
  float ball_dir = (((rand() % 1000000) / 1000000.0f) - 0.5f) * 3.1415926f * 2.0f;
  float ball_dx = cosf(ball_dir);
  float ball_dy = sinf(ball_dir);
  float ball_speed = INITIAL_BALL_SPEED;

  int64_t us = time_us_32();

  pwm_config config = pwm_get_default_config();
  pwm_config_set_wrap(&config, 0x6000);
  pwm_config_set_clkdiv(&config, 2.6f);
  pwm_init(pwm_gpio_to_slice_num(BUZZER_PIN), &config, false);
  pwm_set_gpio_level(BUZZER_PIN, 0x3000);

  gpio_set_function(BUZZER_PIN, GPIO_FUNC_PWM);

  while (true)
  {
    oled.clear();

    for (int i = 0; i < oled.get_width(); i++)
    {
      oled.write(i, 0, 0xFF);
      oled.write(i, oled.get_height() - 1, 0xFF);
    }

    for (int i = 0; i < oled.get_height(); i++)
    {
      oled.write(0, i, 0xFF);
      oled.write(oled.get_width() - 1, i, 0xFF);
    }

    int64_t current = time_us_32();
    float dt = std::abs(current - us) / 1000000.0f;
    us = current;

    int32_t left_enocder_delta = get_encoder_delta(i2c0, LEFT_ENCODER_ADDRESS);
    int32_t right_encoder_delta = get_encoder_delta(i2c0, RIGHT_ENCODER_ADDRESS);

    player.y += 2 * right_encoder_delta;

    float diff = ball.middle_v() - ai.middle_v();
    ai.y += 1.6f * diff * dt;

    ball.x += ball_dx * dt * ball_speed;
    ball.y += ball_dy * dt * ball_speed;

    const rect player_top = { player.left(), player.top(), player.w - 2, 1 };
    const rect player_bottom = { player.left(), player.bottom() - 1, player.w - 2, 1 };
    if (ball.overlaps(player_top) || ball.overlaps(player_bottom))
    {
      ball_dy = -ball_dy;

      beep();
    }
    else if (ball.overlaps(player))
    {
      ball_dx = -ball_dx;
      ball.x = player.right();
      ball_speed += 1.0f;

      float angle = atan2f(ball_dy, ball_dx);
      float angle_delta = (ball.middle_v() - player.middle_v()) / 30.0f;

      ball_dx = cosf(angle + angle_delta);
      ball_dy = sinf(angle + angle_delta);

      beep();
    }

    const rect ai_top = { ai.left() + 2, ai.top(), ai.w - 2, 1 };
    const rect ai_bottom = { ai.left() + 2,  ai.bottom() - 1, ai.w - 2, 1 };
    if (ball.overlaps(ai_top) || ball.overlaps(ai_bottom))
    {
      ball_dy = -ball_dy;

      beep();
    }
    else if (ball.overlaps(ai))
    {
      ball_dx = -ball_dx;
      ball.x = ai.left() - ball.w;
      ball_speed += 1.0f;

      float angle = atan2f(ball_dy, ball_dx);
      // ball second since the paddle is facing the other way
      float angle_delta = (ai.middle_v() - ball.middle_v()) / 30.0f;

      ball_dx = cosf(angle + angle_delta);
      ball_dy = sinf(angle + angle_delta);

      beep();
    }

    if (ball.left() < 0)
    {
      ball_dx = -ball_dx;
      ball.x = 0;

      beep();
    }
    if (ball.right() >= oled.get_width())
    {
      ball_dx = -ball_dx;
      ball.x = oled.get_width() - ball.w;

      beep();
    }

    if (ball.top() < 0)
    {
      ball_dy = -ball_dy;
      ball.y = 0;

      beep();
    }
    if (ball.bottom() >= oled.get_height())
    {
      ball_dy = -ball_dy;
      ball.y = oled.get_height() - ball.h;

      beep();
    }

    player.draw(oled);
    ai.draw(oled);
    ball.draw(oled);

    oled.swap_buffers();
    oled.update();
  }
}
